# -*- coding: utf-8 -*-

'''
ZOOMA api wrapper
'''

import logging
import time
import requests

ZOOMA = 'https://www.ebi.ac.uk/spot/zooma/v2/api'

__all__ = [
    'ZoomaClient'
]

logger = logging.getLogger(__name__)


class ZoomaClient:
    """Simple client to query zooma

    By default (specifying nothing), Zooma will search its available databases 
    containing curated mappings (and that do not include ontology sources), 
    and if nothing is found it will look in the Ontology Lookup Service (OLS) 
    to predict ontology annotations.

    Example:
        >>> z = ZoomaClient()
        >>> r = z.annotate("mus musculus")
        >>> r[0]['semanticTags']
        ['http://purl.obolibrary.org/obo/NCBITaxon_10090']

        >>> r[0]['confidence']
        'HIGH'

        >>> z.besthit("mus musculus")
        'http://purl.obolibrary.org/obo/NCBITaxon_10090'
    """

    def __init__(self, zooma_base=None,
                required=None, preferred=None, ontologies='none'):
        """
        :param zooma_base: An optional, custom URL for the Zooma RESTful API.
        """
        self.base = (zooma_base if zooma_base else ZOOMA).rstrip('/')
        self._annotate = self.base + '/services/annotate'

    def highconfhits(self, name):
        return [m for m in self.annotate(name) if m['confidence'] == 'HIGH']
    
    def besthit(self, name):
        return self.highconfhits(name)[0]['semanticTags'][0]


    @staticmethod
    def _make_filter_string(reqd, prefd, ontos):
        '''
        >>> z = ZoomaClient()
        >>> z._make_filter_string("atlas,gwas","gwas","efo")
        'required:[atlas,gwas],preferred:[gwas],ontologies:[efo]'
        >>> z._make_filter_string(None,None,"none")
        'ontologies:[none]'
        '''
        filters = []
        if reqd:
            filters.append('required:[{}]'.format(reqd))
        if prefd:
            filters.append('preferred:[{}]'.format(prefd))
        if ontos:
            filters.append('ontologies:[{}]'.format(ontos))
        
        return ','.join(filters)


    def annotate(self, name, property_type = None, 
                required=None, preferred=None, ontologies='none'):

        params = {'propertyValue': name}

        if property_type: params['propertyType'] = property_type 

        #The 'ontologies:[none]' parameter will restrain Zooma from looking in
        #the OLS if no annotation was found.
        params['filters'] = self._make_filter_string(required,preferred,ontologies)

        r = requests.get(self._annotate, params=params)
        logger.debug("Request to Zooma annotate API: {} - {}".format(r.status_code,name))
        r.raise_for_status()
        return r.json()





# -*- coding: utf-8 -*-
__all__ = [
    "OnToma",
    "make_uri"
    ]

from ontoma.downloaders import get_omim_to_efo_mappings, get_ot_zooma_to_efo_mappings
from ontoma.ols import OlsClient
from ontoma.zooma import ZoomaClient
from ontoma.oxo import OxoClient

from ontoma.constants import URLS

import requests
import csv
import json
from io import BytesIO, TextIOWrapper
import obonet

import logging
logger = logging.getLogger(__name__)

# if you want to use across classes
# _cache = {"files":None}


def name_to_label_mapping(obonetwork):
    '''
    builds name <=> label lookup dictionaries starting
    from an OBO file
    '''
    id_to_name = {}
    name_to_id = {}
    for nodeid, data in obonetwork.nodes(data=True):
        id_to_name[nodeid] = data['name']
        name_to_id[data['name']] = nodeid 
        if 'synonym' in data:
            for s in data['synonym']:
                #some str.split voodoo is necessary to unpact the synonyms
                # as specified in the obo file
                name_to_id[s.split('\"')[1]] = nodeid
    return id_to_name, name_to_id


def make_uri(ontology_short_form):
    '''
    Transform a short form ontology code in a full URI. 
    Currently works for EFO, HPO, ORDO and MP. 

    Args:
        ontology_short_form: An ontology code in the short format, like 'EFO:0000270'.

    Returns:
        A full URI.

    Example:
        >>> make_uri('EFO:0000270')
        'http://www.ebi.ac.uk/efo/EFO_0000270'
        
        >>> make_uri('HP_0000270')
        'http://purl.obolibrary.org/obo/HP_0000270'
        
        >>> make_uri('http://purl.obolibrary.org/obo/HP_0000270')
        'http://purl.obolibrary.org/obo/HP_0000270'
    '''
    if ontology_short_form.startswith('http'): 
        return ontology_short_form
    ontology_code = ontology_short_form.replace(':',"_")
    if ontology_code.startswith('EFO'):
        return 'http://www.ebi.ac.uk/efo/'+ontology_code
    elif ontology_code.startswith('HP') or ontology_code.startswith('MP') :
        return 'http://purl.obolibrary.org/obo/' + ontology_code
    elif ontology_code.startswith('Orphanet') :
        return 'http://www.orpha.net/ORDO/' + ontology_code
    else:
        logger.error("Could not build an URI. {} not recognized".format(ontology_code))
        raise Exception


class OnToma(object):
    '''Open Targets ontology mapping wrapper

    The output should always be a EFO/OpenTargets ontology URI.

    Example:
        Initialize the class (which will download EFO,OBO and others):
        
        >>> t=OnToma()

        We can now lookup "asthma" and get:

        >>> t.efo_lookup('asthma')
        'EFO:0000270'

        Search by synonyms coming from the OBO file is also supported

        >>> t.efo_lookup('Asthma unspecified')
        'EFO:0000270'

        Reverse lookups uses the get_efo_label() method

        >>> t.get_efo_label('EFO_0000270')
        'asthma'
        >>> t.get_efo_label('EFO:0000270')
        'asthma'


        Similarly, we can now lookup "Phenotypic abnormality" on HP: 

        >>> t.hp_lookup('Phenotypic abnormality')
        'HP:0000118'
        >>> t.hp_lookup('Narrow nasal tip')
        'HP:0011832'

        Lookup in OLS

        >>> t.ols_lookup('asthma')
        'EFO_0000270'

        OMIM code lookup

        >>> t.omim_lookup('230650')
        'http://www.orpha.net/ORDO/Orphanet_354'
        
        >>> t.zooma_lookup('asthma')
        'http://www.ebi.ac.uk/efo/EFO_0000270'


        Searching the ICD9 code for 'other dermatoses' returns EFO's skin disease:

        >>> t.icd9_lookup('696')
        'EFO:0000676'

        There is also a semi-intelligent wrapper, which tries to guess the 
        best matching strategy:
        
        >>> t.find_efo('asthma')
        'http://www.ebi.ac.uk/efo/EFO_0000270'
        >>> t.find_efo('615877',code='OMIM')
        'http://www.orpha.net/ORDO/Orphanet_202948'
    '''

    def __init__(self, efourl = URLS.EFO, 
                        hpurl = URLS.HP):

        self.logger = logging.getLogger(__name__)

        '''Parse the ontology obo files for exact match lookup'''
        #TODO delay download of the OBO files until the class instance is used
        self._efo = obonet.read_obo(efourl)
        self.logger.info('EFO parsed. Size: {} nodes'.format(len(self._efo)))
        self._hp = obonet.read_obo(hpurl)
        self.logger.info('HP parsed. Size: {} nodes'.format(len(self._hp)))

        '''Create name <=> label mappings'''
        self.efo_to_name, self.name_to_efo = name_to_label_mapping(self._efo)
        
        self.hp_to_name, self.name_to_hp = name_to_label_mapping(self._hp)
        

        '''Initialize API clients'''

        self._ols = OlsClient(ontology=['efo'],field_list=['short_form'])
        self._zooma = ZoomaClient()
        self._oxo = OxoClient()

        '''ICD9 <=> EFO mappings '''
        self._icd9_to_efo = self._oxo.make_mappings(input_source = "ICD9CM",
                                                   mapping_target= 'EFO')

        '''OT specific mappings in our github repos'''
        self._zooma_to_efo_map = get_ot_zooma_to_efo_mappings(URLS.ZOOMA_EFO_MAP)
        self._omim_to_efo = get_omim_to_efo_mappings(URLS.OMIM_EFO_MAP)


    def get_efo_label(self, efocode):
        '''Given an EFO code, returns name/label 
        
        Currently based on the EFO OBO file alone.
        '''
        try:
            return self.efo_to_name[efocode.replace('_',':')]
        except KeyError:
            logger.error('EFO ID {} not found'.format(efocode))
            return None

    def zooma_lookup(self, name):
        '''Searches against the EBI Zooma service for an high confidence mapping
        '''
        return self._zooma.besthit(name)

    def otzooma_map_lookup(self, name):
        '''Searches against the curated OpenTargets mapping we submitted to zooma.
        
        These mappings are usually stored on github.
        NOTE: this is not a lookup to zooma API 
        '''
        return self._zooma_to_efo_map[name]

    def icd9_lookup(self, icd9code):
        '''Searches the ICD9CM <=> EFO mappings returned from the OXO API
        '''
        return self._icd9_to_efo[icd9code]

    def omim_lookup(self, omimcode):
        '''Searches our own curated OMIM <=> EFO mappings
        #FIXME assumes the first is the best hit. is this ok?
        '''
        return self._omim_to_efo[omimcode][0]
    
    def ols_lookup(self, name):
        '''Searches the EBI OLS API for a best match from the EFO
        '''
        return self._ols.besthit(name)['short_form']

    def hp_lookup(self, name):
        '''Searches the HP OBO file for a direct match
        '''
        return self.name_to_hp[name]
    
    def efo_lookup(self, name):
        '''Searches the EFO OBO file for a direct match
        '''
        return self.name_to_efo[name]

    def oxo_lookup(self, other_ontology_id, input_source="ICD9CM"):
        '''Searches in the mappings returned from the EBI OXO API. 

        The function should return an EFO code for any given xref, if one
        exists.

        Args: 
            other_ontology_id: the code that should be mapped to EFO
            input_source: an ontology code. Defaults to 'ICD9CM'.
                Available ontologies are listed at https://www.ebi.ac.uk/spot/oxo/api/datasources?fields=preferredPrefix
        
        Returns:
            str: the EFO code
        '''
        return self._oxo.search(ids=[other_ontology_id],
                                input_source=input_source,
                                mapping_target = 'EFO', 
                                distance = 2)


    def find_efo(self, query, code=None):
        '''Finds the most likely EFO code for a given string or ontology code.

        If the code argument is passed, it will attempt to perform an exact match
        amongst the mappings available. 

        If only a string is passed, it will attempt to match it against mappings,
        but will try using the EBI SPOT APIs if no match is found, until a likely
        code is identified

        **TODO** suggestions and fuzzy search should be returned. A specific 
        exception should be crafted and handled. 

        Args:
            query (str): the disease/phenotype to be matched to an EFO code
            code: accepts one of "ICD9CM", "OMIM"
                **TODO** expand to more ontologies
                If a code is passed, it will attempt to find the code in one of our
                curated mapping datasources. Defaults to None.
        
        '''
        if code:
            try:
                return make_uri(self._find_efo_from_code(query, code=code))
            except Exception as e:
                logger.error(e)
                return None
        else:
            return make_uri(self._find_efo_from_string(query))


    def _find_efo_from_code(self, query, code):
        #FIXME need to properly deal with NotFound scenario
        if code == 'OMIM':
            return self.omim_lookup(query)
        if code == 'ICD9CM':
            return self.icd9_lookup(query)
        logger.warning('Could not find EFO for ID: {} in {}'.format(query, code))
        return None
    

    def _find_efo_from_string(self, query):
        '''Searches for a matching EFO code for a given phenotype/disease string

        operations roughly ordered from least expensive to most expensive
        and also from most authorative to least authorative

        1. search exact match to name in EFO (or synonyms)
        2. (search fuzzy match to name in EFO)
        3. search within our open targets zooma mappings
        3. search in OLS
        4. search in Zooma High confidence set

        '''
        if self.efo_lookup(query):
            return self.efo_lookup(query)
        if self.otzooma_map_lookup(query):
            return self.otzooma_map_lookup(query)
        if self.ols_lookup(query):
            return self.ols_lookup(query)
        if self.zooma_lookup(query):
            return self.zooma_lookup(query)
        logger.warning('Could not find EFO for string: {}'.format(query))
        return None




# -*- coding: utf-8 -*-

"""
OLS API wrapper

Original code borrowed from https://github.com/cthoyt/ols-client/blob/master/src/ols_client/client.py

- Removed ontology and term methods. 
- Added details/parameters for all search methods
"""

import logging
import time
import requests

OLS = 'http://www.ebi.ac.uk/ols'
    

__all__ = [
    'OlsClient'
]

logger = logging.getLogger(__name__)

api_suggest = '/api/suggest'
api_search = '/api/search'
api_select = '/api/select'


class OlsClient:
    """Wraps the functions to query the Ontology Lookup Service.
    
    >>> ols = OlsClient()
    >>> ols.search('asthma')[0]['iri']
    'http://purl.obolibrary.org/obo/NCIT_C28397'

    >>> r = ols.search('asthma',ontology=['efo'],query_fields=['synonym'],field_list=['iri','label'])
    >>> r[0]['iri']
    'http://www.ebi.ac.uk/efo/EFO_0004591'

    >>> r = ols.suggest('asthma', ontology=['efo','ordo','hpo'])
    >>> r[0]['autosuggest']
    'asthma'

    >>> r= ols.select('asthma', ontology=['efo','ordo','hpo'],field_list=['iri'])
    >>> r[0]['iri']
    'http://www.ebi.ac.uk/efo/EFO_0000270'

    >>> ols.search('Myofascial Pain Syndrome',ontology=['efo'])[0]['short_form']
    'EFO_1001054'

    >>> [x['short_form'] for x in ols.select('alzheimer')[:2]]
    ['NCIT_C2866', 'NCIT_C38778']

    You can also pass your favourite parameters at class instantiation:

    >>> ot_ols = OlsClient(ontology=['efo'],field_list=['short_form'])
    >>> ot_ols.search('asthma')[0]['short_form']
    'EFO_0000270'
    >>> ot_ols.besthit('asthma')['short_form']
    'EFO_0000270'
    """

    def __init__(self, ols_base=None, ontology=None, field_list=None, query_fields=None):
        """
        :param ols_base: An optional, custom URL for the OLS RESTful API.
        """
        self.base = (ols_base if ols_base else OLS).rstrip('/')

        self.ontology = ontology if ontology else None
        self.field_list = field_list if field_list else None
        self.query_fields = query_fields if query_fields else None

        self.ontology_suggest = self.base + api_suggest
        self.ontology_select = self.base + api_select
        self.ontology_search = self.base + api_search

    def besthit(self, name):
        return self.search(name)[0]

    def search(self, name, query_fields=None, ontology=None, field_list=None):
        """Searches the OLS with the given term

        By default the search is performed over term labels, synonyms, 
        descriptions, identifiers and annotation properties.
        Specify the fields to query, the defaults are 
        {label, synonym, description, short_form, obo_id, annotations, logical_description, iri}

        """
        params = {'q': name}

        if ontology:
            params['ontology'] = ','.join(ontology)
        elif self.ontology:
            params['ontology'] = ','.join(self.ontology)

        if query_fields:
            params['queryFields'] = ','.join(query_fields)
        elif self.query_fields:
            params['queryFields'] = ','.join(self.query_fields)

        if field_list:
            params['fieldList'] = ','.join(field_list)
        elif self.field_list:
            params['fieldList'] = ','.join(self.field_list)

        r = requests.get(self.ontology_search, params=params)
        r.raise_for_status()
        if r.json()['response']['numFound']:
            return r.json()['response']['docs']
        else:
            logger.debug('OLS search returned empty response for {}'.format(name))
            return None


    def suggest(self, name, ontology=None):
        """Suggest terms from an optional list of ontologies

        .. seealso:: https://www.ebi.ac.uk/ols/docs/api#_suggest_term
        """
        params = {'q': name}
        if ontology:
            params['ontology'] = ','.join(ontology)
        r = requests.get(self.ontology_suggest, params=params)
        r.raise_for_status()

        if r.json()['response']['numFound']:
            return r.json()['response']['docs']
        else:
            logger.debug('OLS suggest returned empty response for {}'.format(name))
            return None


    def select(self, name, ontology=None, type=None, field_list=None):
        """Select terms,
        Tuned specifically to support applications such as autocomplete.
        
        .. seealso:: https://www.ebi.ac.uk/ols/docs/api#_select
        """
        params = {'q': name}
        if ontology:
            params['ontology'] = ','.join(ontology)
        if field_list:
            params['fieldList'] = ','.join(field_list)
        r = requests.get(self.ontology_select, params=params)
        r.raise_for_status()

        if r.json()['response']['numFound']:
            return r.json()['response']['docs']
        else:
            logger.debug('OLS select returned empty response for {}'.format(name))
            return None

    


# -*- coding: utf-8 -*-

"""
wrapper for the OXO api 
(reverse engineered! argh!)
"""

import logging
import time
import requests
from requests import HTTPError


__all__ = [
    'OxoClient'
]

OXO = 'https://www.ebi.ac.uk/spot/oxo/api'
MAXSIZE = 500

logger = logging.getLogger(__name__)




class OxoClient:
    ''' OXO wrapper class

    >>> oxo = OxoClient()
    >>> len(oxo._sources)
    940

    >>> first_result = list(oxo.search(input_source="ICD9CM"))[:1][0]
    >>> first_result['curie']
    'ICD9CM:730.92'

    >>> for r in oxo.search(ids=['ICD9CM:171.6'],input_source="ICD9CM"):
    ...     print(r['label'])
    Malignant neoplasm of connective and other soft tissue of pelvis

    >>> icd9s = oxo.make_mappings(input_source="ICD9CM", distance=2)
    >>> icd9s['733.0']
    'EFO:0003882'
    '''

    def __init__(self, base_url=OXO.rstrip('/')):
        self._baseapi = base_url
        self._searchapi = base_url + '/search'
        self._mappingsapi = base_url + '/mappings'
        self._datasourcesapi = base_url + '/datasources'

        self.get_data_sources()


    @staticmethod
    def _pages(method, url, sleep = None, *args, **kwargs):
        '''generator expression to paginate through OXO methods
        '''
        r = method(url,*args,**kwargs)
        r.raise_for_status()
        resp = r.json()
        yield resp
        if 'next' not in resp['_links']:
            return
        while 'next' in resp['_links']:
            if sleep:
                time.sleep(sleep)
            logger.debug(resp['_links']['next']['href'])
            r = method(resp['_links']['next']['href'],*args,**kwargs)
            r.raise_for_status()
            resp = r.json()
            yield resp


    def get_data_sources(self):
        self._sources = set()
        payload = {'size': MAXSIZE}
        for p in self._pages(requests.get, self._datasourcesapi, params=payload):
            self._sources.update([x['prefix'] for x in p['_embedded']['datasources']])
        return


    
    def search(self, ids = None, input_source = None, mapping_target = 'EFO', distance = 1, size = MAXSIZE):
        '''iterates over the mappings, each being a dict with the following keys:
        [
        "_links",
        "curie",
        "label",
        "mappingResponseList",
        "queryId",
        "querySource"
        ]
        '''
        assert input_source in self._sources
        if size > MAXSIZE:
            raise ValueError('Maximum size is {}. Given: {}'.format(MAXSIZE,size))

        payload = {
            "size": size,
            "ids": ids,
            'inputSource': input_source,
            'mappingTarget': mapping_target,
            'distance': str(distance)
        }
        logger.debug('Passing parameters: {}'. format(payload))
        try:
            for page in self._pages(requests.post, self._searchapi, data=payload):
                logger.debug('Returned {} mappings'.format(len(page['_embedded']['searchResults'])))
                for mapping in page['_embedded']['searchResults']:
                    yield mapping
        except HTTPError as e:
            logger.error(e.response.json()['message'])
            return


    
    def make_mappings(self, input_source = "ICD9CM",**kwargs):
        src = self.search(input_source=input_source,**kwargs )
        mappings = {}
        for row in src:
            mappings[ row['curie'].split(':')[1] ] = row['mappingResponseList'][0]['curie']
        return mappings








class URLS:
    '''constants for urls for ontology files, mapping and API'''

    EFO = 'https://github.com/EBISPOT/efo/raw/v2018-01-15/efo.obo'
    HP = 'http://purl.obolibrary.org/obo/hp.obo'

    OMIM_EFO_MAP = 'https://raw.githubusercontent.com/opentargets/platform_semantic/master/resources/xref_mappings/omim_to_efo.txt'
    ZOOMA_EFO_MAP = 'https://raw.githubusercontent.com/opentargets/platform_semantic/master/resources/zooma/cttv_indications_3.txt'


'''
Functions that download mapping flat-files from github repo or other
repositories
'''

__all__ = [
    "get_ot_zooma_to_efo_mappings",
    "get_omim_to_efo_mappings"
    ]

from ontoma.constants import URLS
import csv
import requests
import logging
logger = logging.getLogger(__name__)

def get_omim_to_efo_mappings(url):
    '''returns a dictionary that maps OMIM codes to EFO_uri
    >>> d = get_omim_to_efo_mappings(URLS.OMIM_EFO_MAP)
    >>> d['609909']
    ['http://www.orpha.net/ORDO/Orphanet_154', 'http://www.orpha.net/ORDO/Orphanet_217607']
    '''
    mappings = {}
    logger.debug("OMIM to EFO mappings - requesting from URL {}".format(url))
    with requests.get(url, stream=True) as r:
        for i,row in enumerate(csv.DictReader(r.iter_lines(decode_unicode=True),delimiter='\t')):
            if row['OMIM'] not in mappings:
                mappings[row['OMIM']] = []
            mappings[row['OMIM']].append(row['efo_uri'])
    logger.debug("Parsed {} rows".format(i))
    return mappings


def get_ot_zooma_to_efo_mappings(url):
    '''download zooma and returns a dict
    >>> d = get_ot_zooma_to_efo_mappings(URLS.ZOOMA_EFO_MAP)
    >>> d['skeletal dysplasias']
    'http://purl.obolibrary.org/obo/HP_0002652'
    '''
    mappings = {}
    logger.debug("ZOOMA to EFO mappings - requesting from URL {}".format(url))
    with requests.get(url, stream=True) as r:
        for i,row in enumerate(csv.DictReader(r.iter_lines(decode_unicode=True),delimiter='\t')):
            #(study, bioentity, property_type, property_value, semantic_tag, annotator, annotation_date)
            # Note here should be 1:1 correspondence
            mappings[row['PROPERTY_VALUE'].lower()] = row['SEMANTIC_TAG']
    logger.debug("Parsed {} rows".format(i))
    return mappings


